#include "Brute2.hpp"
#include "Geometry.hpp"
#include "IsValid.hpp"
#include "CalcScore.hpp"
#include "FastPoints.hpp"

using std::vector;
using std::pair;

Brute2::Brute2(const Problem& problem,
               const FastPoints& fastPoints,
               const vector<Point>& points,
               OnSolution onSolution,
               OnDebug onDebug) :
    m_problem(problem),
    m_fastPoints(fastPoints),
    m_points(points),
    m_onSolution(onSolution),
    m_onDebug(onDebug)
{
    m_solution.m_vertices.resize(problem.m_vertices.size());

    m_vertexPoints.resize(problem.m_vertices.size());
    m_backupPoints.resize(problem.m_vertices.size());
    for (auto& backupPoints : m_backupPoints)
    {
        backupPoints.resize(problem.m_vertices.size());
    }

    m_cornerVertices.resize(problem.m_hole.size());
    m_depthCornerVertices.resize(problem.m_vertices.size());
    m_cornerProcessed.resize(problem.m_hole.size());
}

Brute2::~Brute2()
{
}

bool Brute2::solve(uint64_t maxNodes,
                   size_t firstVertex,
                   const vector<Point>* pTargets)
{
    m_pTargets = pTargets;
    m_maxNodes = maxNodes;
    m_nodeCount = 0;
    for (auto& vertex : m_solution.m_vertices)
    {
        vertex = Point(-999, -999);
    }
    m_processed.clear();
    m_processed.resize(m_problem.m_vertices.size());

    for (auto& vertexPoints : m_vertexPoints)
    {
        //vertexPoints.clear();
        vertexPoints = m_points;
    }

    m_firstVertex = firstVertex;

    //m_vertexPoints[3].clear();
    //m_vertexPoints[3].push_back(m_problem.m_hole[4]);
    //m_vertexPoints[9].clear();
    //m_vertexPoints[9].push_back(m_problem.m_hole[8]);
    auto place = [&](size_t iCorner, size_t iVertex)
    {
        m_vertexPoints[iVertex].resize(1);
        m_vertexPoints[iVertex][0] = m_problem.m_hole[iCorner];
    };
    (void)place;

#if 0
    place(7, 50);
    place(6, 49);
    place(12, 34);
    place(13, 60);
    place(14, 55);
    place(15, 52);
    place(16, 65);
    place(17, 80);
    place(18, 78);
    place(19, 83);
    place(30, 69);
    place(31, 93);
    place(36, 105);
    place(38, 91);
    place(39, 72);
    place(44, 75);
    place(50, 53);
    place(52, 39);
    place(53, 20);
    place(54, 54);
    place(56, 58);
    //place(60, 47);
    place(61, 21);
    place(62, 14);
    place(63, 24);
    place(64, 7);
    //place(24, 94);
    m_firstVertex = 75;
#endif
#if 0
    // Problem 4
    place(10, 2);
    place(7, 31);
#endif
#if 0
    // Problem 122
    place(3, 8);
    place(4, 19);
    place(5, 35);
    place(6, 45);

    place(14, 4);
    place(15, 6);
    place(16, 2);
    place(17, 9);
    place(18, 27);
    place(19, 18);

    place(42, 210);
    place(43, 209);
    place(44, 213);
    place(45, 214);
    place(46, 212);
    place(47, 208);
    place(48, 188);
    place(49, 182);
    place(50, 164);

    place(59, 79);
    place(60, 56);
    place(61, 31);
    place(62, 14);
    place(63, 13);
#endif
#if 0
    // Problem 64
    place(12, 39);
    place(15, 34);
    place(16, 55);
    place(17, 53);
    place(18, 79);
    place(19, 80);
    place(21, 68);
    //place(23, 10);
    place(39, 64);
#endif
#if 0
    // Problem 85
    place(21, 94);
    place(31, 112);
    place(41, 98);
    place(42, 106);
    place(50, 75);
#endif

    return recurse(m_solution.m_vertices.size(), (size_t)-1);
}

bool Brute2::recurse(size_t depth, size_t lastVertex)
{
    //printf("> %" PRIuZ "\n", depth);
    m_nodeCount++;
    if ((m_nodeCount & 0xff) == 0)
    {
        if (m_onDebug)
        {
            if (m_onDebug(m_solution))
            {
                return true;
            }
        }
    }

    if (m_maxNodes != 0 && m_nodeCount >= m_maxNodes)
    {
        return true;
    }

    if (depth == 0)
    {
#if 1
        if (!isValid(m_problem, m_solution))
        {
            printf("Not valid!\n");
            exit(1);
        }
#endif
        uint64_t score = calcScore(m_problem, m_solution);
        //printf("Found node count = %" PRIu64 "\n", m_nodeCount);
        //printf("< %" PRIuZ "\n", depth);
        return m_onSolution(m_solution, score);
    }

    size_t bestVertex = (size_t)-1;
    size_t bestCorner = (size_t)-1;
    size_t bestOkCount = (size_t)-1;
    if (lastVertex == (size_t)-1)
    {
#if 0
        bestVertex = m_firstVertex;
        bestCorner = (size_t)-1;
        bestOkCount = m_points.size();
#endif
#if 1
        bestVertex = (size_t)-1;
        bestCorner = 0;
        bestOkCount = m_problem.m_vertices.size();
#endif
    }
    else
    {
        for (size_t iCorner = 0; iCorner < m_problem.m_hole.size(); iCorner++)
        {
            m_cornerVertices[iCorner].clear();
            m_cornerProcessed[iCorner] = 0;
        }

        for (size_t iVertex = 0; iVertex < m_problem.m_vertices.size(); iVertex++)
        {
            if (!m_processed[iVertex]) continue;
            Point point = m_solution.m_vertices[iVertex];
            if (m_fastPoints.isCorner(point))
            {
                for (size_t iCorner = 0; iCorner < m_problem.m_hole.size(); iCorner++)
                {
                    if (m_problem.m_hole[iCorner] == point)
                    {
                        m_cornerProcessed[iCorner] = 1;
                        break;
                    }
                }
            }
        }

        for (size_t iVertex = 0; iVertex < m_problem.m_vertices.size(); iVertex++)
        {
            if (m_processed[iVertex]) continue;
            auto& vertexPoints = m_vertexPoints[iVertex];
            auto& backupPoints = m_backupPoints[depth - 1][iVertex];
            vertexPoints.swap(backupPoints);
            vertexPoints.clear();
            if (bestOkCount == 0) continue;

            size_t edgeIdx = (size_t)-1;
            uint64_t dOld = 0;
            for (auto& neighbor : m_fastPoints.getNeighbors(lastVertex))
            {
                if (neighbor.m_vertexIdx == iVertex)
                {
                    edgeIdx = neighbor.m_edgeIdx;
                    dOld = m_fastPoints.getEdgeInfo(edgeIdx).m_d2;
                    break;
                }
            }

            uint64_t epsilon = m_problem.m_epsilon;
            for (auto point : backupPoints)
            {
                if (edgeIdx != (size_t)-1)
                {
                    uint64_t dNew = calcD(m_solution.m_vertices[lastVertex],
                                          point);
                    if (!isValidD(dOld, dNew, epsilon))
                    {
                        continue;
                    }
                    if (!edgeInPolygon(m_problem.m_hole,
                                       m_solution.m_vertices[lastVertex],
                                       point))
                    {
                        continue;
                    }
                }

#if 1
                {
                    uint64_t maxD1_1000 =
                        m_fastPoints.getShortestPath(iVertex, lastVertex);
                    uint64_t d2 = calcD(point, m_solution.m_vertices[lastVertex]);
                    if (maxD1_1000 * maxD1_1000 < d2 * 1000000)
                    {
                        continue;
                    }
                }
#endif

                vertexPoints.push_back(point);

                if (m_fastPoints.isCorner(point))
                {
                    for (size_t iCorner = 0; iCorner < m_problem.m_hole.size(); iCorner++)
                    {
                        if (m_cornerProcessed[iCorner]) continue;
                        if (m_problem.m_hole[iCorner] == point)
                        {
                            m_cornerVertices[iCorner].push_back(iVertex);
                            break;
                        }
                    }
                }
            }

            if (vertexPoints.empty())
            {
                bestVertex = iVertex;
                bestCorner = (size_t)-1;
                bestOkCount = 0;
            }
        }

#if 0
        for (size_t iVertex = 0; iVertex < m_problem.m_vertices.size(); iVertex++)
        {
            if (m_processed[iVertex]) continue;
            printf("%2" PRIuZ ": %" PRIuZ "\n", iVertex, m_vertexPoints[iVertex].size());
        }
#endif

        if (bestOkCount != 0)
        {
            for (size_t iVertex = 0; iVertex < m_problem.m_vertices.size(); iVertex++)
            {
                if (m_processed[iVertex]) continue;
                auto& vertexPoints = m_vertexPoints[iVertex];
                if (vertexPoints.empty()) continue;
                size_t okCount = vertexPoints.size();

                if (okCount < bestOkCount)
                {
                    bestVertex = iVertex;
                    bestCorner = (size_t)-1;
                    bestOkCount = okCount;
                }
            }

            for (size_t iCorner = 0; iCorner < m_problem.m_hole.size(); iCorner++)
            {
                if (m_cornerProcessed[iCorner]) continue;
                size_t okCount = m_cornerVertices[iCorner].size();
                if (okCount < bestOkCount)
                {
                    bestVertex = (size_t)-1;
                    bestCorner = iCorner;
                    bestOkCount = okCount;
                }
            }
        }
    }

#if 0
    printf("bestOkCount = %" PRIuZ "\n", bestOkCount);
    if (bestVertex != (size_t)-1)
    {
        printf("bestVertex = %" PRIi64 "\n", bestVertex);
    }
    if (bestCorner != (size_t)-1)
    {
        printf("bestCorner = %" PRIi64 "\n", bestCorner);
    }
    printf("\n");
#endif
    //if (bestOkCount == 0) return false;

    if (bestOkCount != 0 && bestVertex != (size_t)-1)
    {
        auto& points =
            //lastVertex == (size_t)-1 ?
            //m_points :
            m_vertexPoints[bestVertex];
        size_t numPoints = points.size();

#if 1
        vector<pair<uint64_t, Point>> greedyPoints(numPoints);
        if (m_pTargets)
        {
            Point target = (*m_pTargets)[bestVertex];
            if (target.x == -999)
            {
                size_t numCorners = m_problem.m_hole.size();
                vector<uint64_t> corners(numCorners);
                for (size_t iCorner = 0; iCorner < numCorners; iCorner++)
                {
                    uint64_t minD = std::numeric_limits<uint64_t>::max();
                    for (size_t iVertex = 0; iVertex < m_problem.m_vertices.size(); iVertex++)
                    {
                        if (!m_processed[iVertex]) continue;
                        minD = std::min(minD, calcD(m_problem.m_hole[iCorner],
                                                    m_solution.m_vertices[iVertex]));
                    }
                    corners[iCorner] = minD;
                }

                for (size_t iPoint = 0; iPoint < numPoints; iPoint++)
                {
                    Point point = points[iPoint];
                    uint64_t score = 0;
                    for (size_t iCorner = 0; iCorner < numCorners; iCorner++)
                    {
                        uint64_t minD = corners[iCorner];
                        minD = std::min(minD, calcD(m_problem.m_hole[iCorner],
                                                    point));
                        score += minD;
                    }
                    greedyPoints[iPoint].first = score;
                    greedyPoints[iPoint].second = point;
                }
            }
            else
            {
                for (size_t iPoint = 0; iPoint < numPoints; iPoint++)
                {
                    Point point = points[iPoint];
                    uint64_t score = calcD(target, point);
                    greedyPoints[iPoint].first = score;
                    greedyPoints[iPoint].second = point;
                }
            }
        }
        else
        {
            for (size_t iPoint = 0; iPoint < numPoints; iPoint++)
            {
                Point point = points[iPoint];
                uint64_t score = 0;
                greedyPoints[iPoint].first = score;
                greedyPoints[iPoint].second = point;
            }
        }
        std::sort(greedyPoints.begin(), greedyPoints.end());
#endif

        m_processed[bestVertex] = 1;
        for (size_t iPoint = 0; iPoint < numPoints; iPoint++)
        {
#if 0
            if (iPoint == 3) break;
#endif
            Point point = greedyPoints[iPoint].second;
            m_solution.m_vertices[bestVertex] = point;
            if (recurse(depth - 1, bestVertex))
            {
                return true;
            }
            m_solution.m_vertices[bestVertex] = Point(-999, -999);
        }
        m_processed[bestVertex] = 0;
    }
    else if (bestOkCount != 0 && bestCorner != (size_t)-1)
    {
        auto& cornerVertices = m_depthCornerVertices[depth - 1];
        if (lastVertex == (size_t)-1)
        {
            cornerVertices.resize(m_problem.m_vertices.size());
            for (size_t iVertex = 0; iVertex < m_problem.m_vertices.size(); iVertex++)
            {
                cornerVertices[iVertex] = iVertex;
            }
        }
        else
        {
            cornerVertices = m_cornerVertices[bestCorner];
        }

        for (size_t iVertex : cornerVertices)
        {
            m_processed[iVertex] = 1;
            Point point = m_problem.m_hole[bestCorner];
            m_solution.m_vertices[iVertex] = point;
            if (recurse(depth - 1, iVertex))
            {
                return true;
            }
            m_solution.m_vertices[iVertex] = Point(-999, -999);
            m_processed[iVertex] = 0;
        }
    }

    if (lastVertex != (size_t)-1)
    {
        for (size_t iVertex = 0; iVertex < m_problem.m_vertices.size(); iVertex++)
        {
            if (m_processed[iVertex]) continue;
            auto& vertexPoints = m_vertexPoints[iVertex];
            auto& backupPoints = m_backupPoints[depth - 1][iVertex];
            vertexPoints.clear();
            vertexPoints.swap(backupPoints);
        }
    }

    //printf("< %" PRIuZ "\n", depth);
    return false;
}
