#ifndef BRUTE2_HPP
#define BRUTE2_HPP

#include "Common.hpp"
#include "Problem.hpp"

class FastPoints;

typedef std::function<bool(const Solution& solution,
                           uint64_t score)> OnSolution;

typedef std::function<bool(const Solution& solution)> OnDebug;

class Brute2
{
public:
    Brute2(const Problem& problem,
           const FastPoints& fastPoints,
           const std::vector<Point>& points,
           OnSolution onSolution,
           OnDebug onDebug);
    ~Brute2();

    bool solve(uint64_t maxNodes = 0,
               size_t firstVertex = 0,
               const std::vector<Point>* pTargets = nullptr);

private:
    bool recurse(size_t depth, size_t lastVertex);

private:
    const Problem& m_problem;
    const FastPoints& m_fastPoints;
    const std::vector<Point>& m_points;
    const std::vector<Point>* m_pTargets = nullptr;
    std::vector<std::vector<Point>> m_vertexPoints;
    std::vector<std::vector<std::vector<Point>>> m_backupPoints;
    std::vector<Point> m_pointsTmp;
    std::vector<std::vector<size_t>> m_cornerVertices;
    std::vector<std::vector<size_t>> m_depthCornerVertices;
    size_t m_firstVertex = 0;
    OnSolution m_onSolution;
    OnDebug m_onDebug;
    uint64_t m_maxNodes = 0;
    uint64_t m_nodeCount = 0;
    Solution m_solution;
    std::vector<uint8_t> m_processed;
    std::vector<uint8_t> m_cornerProcessed;
};

#endif
