#ifndef FASTPOINTS_HPP
#define FASTPOINTS_HPP

#include "Common.hpp"
#include "Point.hpp"

class Problem;

class EdgeInfo
{
public:
    uint64_t m_d2;
    uint64_t m_maxParallelD1;
    uint64_t m_maxD1_1000;
    bool m_duplicate;
};

class DistInfo
{
public:
    uint64_t m_dist;
    bool m_zero;
    std::vector<int64_t> m_parallel;
    std::vector<int64_t> m_diagonal;
    std::vector<Point> m_between;
};

class Neighbor
{
public:
    size_t m_edgeIdx;
    size_t m_vertexIdx;
};

class FastPoints
{
public:
    FastPoints();
    ~FastPoints();

    void init(const Problem& problem,
              const std::vector<Point>& points);
    void getPoints(size_t edgeIdx,
                   Point point,
                   std::vector<Point>* pPoints) const;

    const EdgeInfo& getEdgeInfo(size_t edgeIdx) const
    {
        return m_edgeInfo[edgeIdx];
    }

    bool isInGrid(Point point) const
    {
        return
            point.y >= m_minY &&
            point.y <= m_maxY &&
            point.x >= m_minX &&
            point.x <= m_maxX &&
            ((m_grid[(point.y - m_minY) * m_stride + (point.x - m_minX)] & 1) != 0);
    }
    bool isCorner(Point point) const
    {
        return
            point.y >= m_minY &&
            point.y <= m_maxY &&
            point.x >= m_minX &&
            point.x <= m_maxX &&
            ((m_grid[(point.y - m_minY) * m_stride + (point.x - m_minX)] & 2) != 0);
    }

    uint64_t getShortestPath(size_t iVertex, size_t jVertex) const
    {
        return m_shortestPaths[iVertex * m_numVertices + jVertex];
    }

    const std::vector<Neighbor>& getNeighbors(size_t iVertex) const
    {
        return m_neighbors[iVertex];
    }

public:
    int64_t m_minX = 0;
    int64_t m_maxX = 0;
    int64_t m_minY = 0;
    int64_t m_maxY = 0;
private:
    int64_t m_stride = 0;
    size_t m_numVertices = 0;
    std::vector<uint8_t> m_grid;
    std::vector<EdgeInfo> m_edgeInfo;
    std::vector<size_t> m_edgeMap;
    std::vector<DistInfo> m_distInfo;
    std::vector<uint64_t> m_shortestPaths;
    std::vector<std::vector<Neighbor>> m_neighbors;
};

#endif
