#include "Geometry.hpp"
#include <limits>

using std::vector;
using std::pair;

int64_t getOrientation(const vector<Point>& polygon)
{
    int64_t orientation = 0;
    Point prevVertex = polygon.back();
    for (Point vertex : polygon)
    {
        orientation += prevVertex.x * vertex.y;
        orientation -= vertex.x * prevVertex.y;
        prevVertex = vertex;
    }
    return orientation;
}

namespace
{
    bool pointInPolygonImpl(const vector<Point>& polygon, Point point, bool on)
    {
        uint64_t leftCount = 0;
        uint64_t rightCount = 0;
        int64_t x = point.x;
        int64_t y = point.y;
        Point v0 = polygon.back();
        for (Point v1 : polygon)
        {
            int64_t x0 = v0.x;
            int64_t y0 = v0.y;
            int64_t x1 = v1.x;
            int64_t y1 = v1.y;
            if (x0 == x && y0 == y)
            {
                return true;
            }
            else if (x1 == x && y1 == y)
            {
                return true;
            }
            else if (y0 < y && y1 < y)
            {
            }
            else if (y0 > y && y1 > y)
            {
            }
            else if (y0 == y && y1 == y)
            {
                if (x >= x0 && x <= x1) return true;
                if (x >= x1 && x <= x0) return true;
            }
            else if (y0 == y)
            {
                if (y1 < y)
                {
                    if (x0 < x) { leftCount++; } else { rightCount++; }
                }
            }
            else if (y1 == y)
            {
                if (y0 < y)
                {
                    if (x1 < x) { leftCount++; } else { rightCount++; }
                }
            }
            else
            {
                int64_t cmp;
                if (y0 < y1)
                {
                    cmp = (x0 - x) * (y1 - y0) + (x1 - x0) * (y - y0);
                }
                else // if (y1 < y0)
                {
                    cmp = (x1 - x) * (y0 - y1) + (x0 - x1) * (y - y1);
                }
                if (cmp < 0) { leftCount++; }
                else if (cmp > 0) { rightCount++; }
                else { return true; }
            }
            v0 = v1;
        }
        if (on)
        {
            return false;
        }
        else
        {
            return (leftCount % 2) == 1;
        }
    }

    vector<Point> getPointsInPolygonImpl(const vector<Point>& polygon, bool on)
    {
        int64_t minX = std::numeric_limits<int64_t>::max();
        int64_t maxX = std::numeric_limits<int64_t>::min();
        int64_t minY = std::numeric_limits<int64_t>::max();
        int64_t maxY = std::numeric_limits<int64_t>::min();
        for (Point p : polygon)
        {
            minX = std::min(minX, p.x);
            maxX = std::max(maxX, p.x);
            minY = std::min(minY, p.y);
            maxY = std::max(maxY, p.y);
        }

        vector<Point> points;

        for (int64_t y = minY; y <= maxY; y++)
        {
            for (int64_t x = minX; x <= maxX; x++)
            {
                Point point(x, y);
                if (pointInPolygonImpl(polygon, point, on))
                {
                    points.push_back(point);
                }
            }
        }

        return points;
    }
}

bool pointInPolygon(const vector<Point>& polygon, Point point)
{
    return pointInPolygonImpl(polygon, point, false);
}

vector<Point> getPointsInPolygon(const vector<Point>& polygon)
{
    return getPointsInPolygonImpl(polygon, false);
}

bool pointOnPolygon(const vector<Point>& polygon, Point point)
{
    return pointInPolygonImpl(polygon, point, true);
}

vector<Point> getPointsOnPolygon(const vector<Point>& polygon)
{
    return getPointsInPolygonImpl(polygon, true);
}

class Rat
{
public:
    Rat() {}
    Rat(int64_t n_, int64_t d_) : n(d_ < 0 ? -n_ : n_), d(d_ < 0 ? -d_ : d_)
    {
        if (d == 0)
        {
            fprintf(stderr, "Denominator is zero\n");
            exit(1);
        }
    }

    bool operator==(const Rat& other) const
    {
        return n * other.d == other.n * d;
    }
    bool operator!=(const Rat& other) const { return !(*this == other); }
    bool operator<(const Rat& other) const
    {
        return n * other.d < other.n * d;
    }
    bool operator>(const Rat& other) const { return other < *this; }
    bool operator<=(const Rat& other) const { return !(other < *this); }
    bool operator>=(const Rat& other) const { return !(*this < other); }

    int64_t n;
    int64_t d;
};

namespace
{
    // Determine which side of the line AB the point C lies on
    int64_t getSide(Point a, Point b, Point c)
    {
        Point ab = b - a;
        Point perp = Point(ab.y, -ab.x);
        Point ac = c - a;
        return dot(perp, ac);
    }
}

bool edgeInPolygon(const vector<Point>& polygon, Point pa, Point pb)
{
#if 0
    vector<Rat> crossings;
    vector<pair<Rat, Rat>> ranges;
    crossings.reserve(polygon.size());
    ranges.reserve(polygon.size());
#endif
    Rat crossings[108];
    pair<Rat, Rat> ranges[108];
    size_t numCrossings = 0;
    size_t numRanges = 0;

    Point v0 = polygon.back();
    for (Point v1 : polygon)
    {
        int64_t side0 = getSide(pa, pb, v0);
        int64_t side1 = getSide(pa, pb, v1);
        Rat t0;
        Rat t1;
        if (pb.x == pa.x)
        {
            t0 = Rat(v0.y - pa.y, pb.y - pa.y);
            t1 = Rat(v1.y - pa.y, pb.y - pa.y);
        }
        else
        {
            t0 = Rat(v0.x - pa.x, pb.x - pa.x);
            t1 = Rat(v1.x - pa.x, pb.x - pa.x);
        }
        if (side0 < 0 && side1 < 0)
        {
        }
        else if (side0 > 0 && side1 > 0)
        {
        }
        else if (side0 == 0 && side1 == 0)
        {
            ranges[numRanges].first = std::min(t0, t1);
            ranges[numRanges].second = std::max(t0, t1);
            numRanges++;
        }
        else if (side0 == 0)
        {
            if (side1 < 0)
            {
                crossings[numCrossings++] = t0;
            }
        }
        else if (side1 == 0)
        {
            if (side0 < 0)
            {
                crossings[numCrossings++] = t1;
            }
        }
        else
        {
            int64_t xd = pb.x - pa.x;
            int64_t yd = pb.y - pa.y;
            int64_t zd = v1.x - v0.x;
            int64_t wd = v1.y - v0.y;
            int64_t xg = v0.x - pa.x;
            int64_t yg = v0.y - pa.y;
            Rat t(xg * wd - yg * zd, xd * wd - yd * zd);
            crossings[numCrossings++] = t;
        }
        v0 = v1;
    }

    std::sort(&crossings[0], &crossings[numCrossings]);
    std::sort(&ranges[0], &ranges[numRanges]);

    size_t iCrossing = 0;
    size_t iRange = 0;
    bool inside = false;
    while (iCrossing < numCrossings &&
           crossings[iCrossing] <= Rat(0, 1))
    {
        inside = !inside;
        iCrossing++;
    }

    Rat t(0, 1);
    while (true)
    {
        if (t >= Rat(1, 1))
        {
            return true;
        }

        if (iCrossing < numCrossings &&
            crossings[iCrossing] <= t)
        {
            inside = !inside;
            iCrossing++;
            continue;
        }

        if (inside)
        {
            if (iCrossing < numCrossings)
            {
                inside = !inside;
                t = crossings[iCrossing];
                iCrossing++;
                continue;
            }
            else
            {
                return true;
            }
        }

        if (iRange < numRanges)
        {
            if (t >= ranges[iRange].second)
            {
                iRange++;
                continue;
            }
            else if (t >= ranges[iRange].first)
            {
                t = ranges[iRange].second;
                iRange++;
                continue;
            }
        }

        return false;
    }
}
