#include "IsValid.hpp"
#include "Problem.hpp"
#include "Geometry.hpp"

bool isValid(const Problem& problem,
             const Solution& solution)
{
    uint64_t numVertices = problem.m_vertices.size();
    if (solution.m_vertices.size() != numVertices)
    {
        return false;
    }

    for (Edge edge : problem.m_edges)
    {
        Point paOld = problem.m_vertices[edge.a];
        Point pbOld = problem.m_vertices[edge.b];
        Point paNew = solution.m_vertices[edge.a];
        Point pbNew = solution.m_vertices[edge.b];
        uint64_t dOld = calcD(paOld, pbOld);
        uint64_t dNew = calcD(paNew, pbNew);

        if (!isValidD(dOld, dNew, problem.m_epsilon))
        {
            return false;
        }
    }

    for (Edge edge : problem.m_edges)
    {
        Point paNew = solution.m_vertices[edge.a];
        Point pbNew = solution.m_vertices[edge.b];

        if (!edgeInPolygon(problem.m_hole, paNew, pbNew))
        {
            return false;
        }
    }

#if 0
    for (Point vertex : solution.m_vertices)
    {
        if (!pointInPolygon(problem.m_hole, vertex))
        {
            return false;
        }
    }
#endif

    return true;
}
