#ifndef POINT_HPP
#define POINT_HPP

#include "Common.hpp"

class Point
{
public:
    Point() : x(0), y(0) {}
    Point(int64_t x_, int64_t y_) : x(x_), y(y_) {}

    Point& operator+=(const Point& other)
    {
        x += other.x;
        y += other.y;
        return *this;
    }

    Point& operator-=(const Point& other)
    {
        x -= other.x;
        y -= other.y;
        return *this;
    }

    int64_t x;
    int64_t y;
};

inline bool operator==(Point a, Point b)
{
    return a.x == b.x && a.y == b.y;
}

inline bool operator!=(Point a, Point b)
{
    return a.x != b.x || a.y != b.y;
}

inline bool operator<(Point a, Point b)
{
    if (a.y != b.y) return a.y < b.y;
    return a.x < b.x;
}

inline bool operator>(Point a, Point b)
{
    return b < a;
}

inline bool operator<=(Point a, Point b)
{
    return !(b < a);
}

inline bool operator>=(Point a, Point b)
{
    return !(a < b);
}

inline Point operator+(Point a, Point b)
{
    return Point(a.x + b.x, a.y + b.y);
}

inline Point operator-(Point a, Point b)
{
    return Point(a.x - b.x, a.y - b.y);
}

inline int64_t dot(Point a, Point b)
{
    return a.x * b.x + a.y * b.y;
}

#endif
