#ifndef PROBLEM_HPP
#define PROBLEM_HPP

#include "Common.hpp"
#include "Point.hpp"

class Edge
{
public:
    uint64_t a = 0;
    uint64_t b = 0;
};

enum class BonusType : uint32_t
{
    Globalist,
    BreakALeg,
    WallHack,
    Superflex
};

class Bonus
{
public:
    Point m_position;
    BonusType m_bonusType = BonusType::Globalist;
    uint64_t m_id = 0;
};

class Problem
{
public:
    std::vector<Point> m_hole;
    std::vector<Point> m_vertices;
    std::vector<Edge> m_edges;
    std::vector<Bonus> m_bonuses;
    uint64_t m_epsilon = 0;
};

class Solution
{
public:
    std::vector<Point> m_vertices;
};

bool readProblem(uint64_t id,
                 std::string* pProblemData);
bool writeProblem(uint64_t id,
                  const std::string& problemData);

bool parseProblem(const std::string& problemData,
                  Problem* pProblem,
                  std::string* pMsg = nullptr);
bool parseProblem(uint64_t id,
                  Problem* pProblem,
                  std::string* pMsg = nullptr);
void formatProblem(const Problem& problem,
                   std::string* pProblemData);

bool readSolution(uint64_t id,
                  std::string* pSolutionData);
bool writeSolution(uint64_t id,
                   const std::string& solutionData);

bool parseSolution(const std::string& solutionData,
                   Solution* pSolution,
                   std::string* pMsg = nullptr);
bool parseSolution(uint64_t id,
                   Solution* pSolution,
                   std::string* pMsg = nullptr);
void formatSolution(const Solution& solution,
                    std::string* pSolutionData);

#endif
