#include "Protocol.hpp"
#include "Http.hpp"
#include "FileUtils.hpp"
#include "StringUtils.hpp"

using std::string;
using std::vector;

namespace Protocol
{
    void init()
    {
        Http::init();
    }

    void cleanup()
    {
        Http::cleanup();
    }

    bool getAPIKey(string* pKey)
    {
        string tmpKey;
        if (!readFile("../../../apikey.txt", &tmpKey))
        {
            return false;
        }

        string key;
        for (char ch : tmpKey)
        {
            if (ch >= 33 && ch <= 126)
            {
                key.push_back(ch);
            }
        }
        if (key.empty())
        {
            return false;
        }

        if (pKey) *pKey = std::move(key);
        return true;
    }

    bool getAPIKeyHeader(vector<string>* pHeaders,
                         std::string* pMsg = nullptr)
    {
        string key;
        if (!getAPIKey(&key))
        {
            if (pMsg) *pMsg = "Error reading API key";
            return false;
        }

        pHeaders->push_back("Authorization: Bearer " + key);
        return true;
    }

    bool getCookie(string* pCookie)
    {
        string tmpCookie;
        if (!readFile("../../../cookie.txt", &tmpCookie))
        {
            return false;
        }

        string cookie;
        for (char ch : tmpCookie)
        {
            if (ch >= 32 && ch <= 126)
            {
                cookie.push_back(ch);
            }
        }
        if (cookie.empty())
        {
            return false;
        }

        if (pCookie) *pCookie = std::move(cookie);
        return true;
    }

    bool getCookieHeader(vector<string>* pHeaders,
                         std::string* pMsg = nullptr)
    {
        string cookie;
        if (!getCookie(&cookie))
        {
            if (pMsg) *pMsg = "Error reading cookie";
            return false;
        }

        pHeaders->push_back(cookie);
        return true;
    }

    bool hello(string* pHelloData,
               string* pMsg)
    {
        vector<string> headers;
        if (!getAPIKeyHeader(&headers, pMsg)) return false;

        string url = "https://poses.live/api/hello";

        string response;
        if (!Http::get(url, &headers, &response, pMsg))
        {
            return false;
        }

        if (pHelloData) *pHelloData = std::move(response);
        return true;
    }

    bool getProblem(uint64_t id,
                    string* pProblemData,
                    string* pMsg)
    {
        vector<string> headers;
        if (!getAPIKeyHeader(&headers, pMsg)) return false;

        string url = strprintf("https://poses.live/api/problems/%" PRIu64 "", id);

        string response;
        if (!Http::get(url, &headers, &response, pMsg))
        {
            return false;
        }

        if (pProblemData) *pProblemData = std::move(response);
        return true;
    }

    bool submitSolution(uint64_t id,
                        const string& solutionData,
                        string* pResponseData,
                        string* pMsg)
    {
        vector<string> headers;
        if (!getAPIKeyHeader(&headers, pMsg)) return false;

        string url = strprintf("https://poses.live/api/problems/%" PRIu64 "/solutions", id);

        string response;
        if (!Http::post(url, solutionData, "application/json", &headers, &response, pMsg))
        {
            return false;
        }

        if (pResponseData) *pResponseData = std::move(response);
        return true;
    }

    bool getScores(string* pScoresData,
                   string* pMsg)
    {
        vector<string> headers;
        if (!getCookieHeader(&headers, pMsg)) return false;

        string url = "https://poses.live/problems";

        string response;
        if (!Http::get(url, &headers, &response, pMsg))
        {
            return false;
        }

        if (pScoresData) *pScoresData = std::move(response);
        return true;
    }
}
