#include "Scores.hpp"
#include "FileUtils.hpp"
#include "ParseUtils.hpp"

using std::string;
using std::vector;

bool readScores(string* pScoresData)
{
    string fileName = "../data/best";
    return readFile(fileName, pScoresData);
}

bool writeScores(const string& scoresData)
{
    string fileName = "../data/best";
    return writeFile(fileName, scoresData);
}

bool parseScores(const string& scoresData,
                 vector<ScoreRec>* pScores,
                 string* pMsg)
{
    size_t start = scoresData.find("<table>");
    if (start == string::npos)
    {
        if (pMsg) *pMsg = "<table> not found";
        return false;
    }
    start += 7;

    size_t end = scoresData.find("</table>", start);
    if (end == string::npos)
    {
        if (pMsg) *pMsg = "</table> not found";
        return false;
    }

    size_t pos = start;
    bool ok = true;

    auto matchChar = [&](char ch)
    {
        if (!ok) return;
        if (pos == end) { ok = false; return; }
        if (scoresData[pos] != ch) { ok = false; return; }
        pos++;
    };

    auto match = [&](const string& str)
    {
        for (auto ch : str)
        {
            matchChar(ch);
        }
    };

    auto skipUntil = [&](char ch)
    {
        while (ok && pos < end && scoresData[pos] != ch)
        {
            pos++;
        }
    };

    auto extractUntil = [&](char ch) -> string
    {
        string str;
        while (ok && pos < end && scoresData[pos] != ch)
        {
            str += scoresData[pos];
            pos++;
        }
        return str;
    };

    match("<tr><th>problem</th><th>your dislikes</th><th>minimal dislikes</th></tr>");

    vector<ScoreRec> scores;

    while (ok && pos < end)
    {
        match("<tr><td><a href=\"");
        skipUntil('"');
        match("\">");
        string strId = extractUntil('<');
        match("</a></td><td>");
        string strScore = extractUntil('<');
        match("</td><td>");
        string strBestScore = extractUntil('<');
        match("</td></tr>");

        if (!ok) break;

        uint64_t id = 0;
        uint64_t score = (uint64_t)-1;
        uint64_t bestScore = (uint64_t)-1;

        if (!parseU64(strId, &id) || id < 1 || id > 1000000)
        {
            if (pMsg) *pMsg = "Bad ID";
            return false;
        }

        if (!strScore.empty())
        {
            if (!parseU64(strScore, &score) || score < 0 || score > 1000000000)
            {
                if (pMsg) *pMsg = "Bad score";
                return false;
            }
        }

        if (!strBestScore.empty())
        {
            if (!parseU64(strBestScore, &bestScore) || bestScore < 0 || bestScore > 1000000000)
            {
                if (pMsg) *pMsg = "Bad bestscore";
                return false;
            }
        }

        auto& scoreRec = scores.emplace_back();
        if (id != scores.size())
        {
            if (pMsg) *pMsg = "Unexpected ID";
            return false;
        }

        scoreRec.m_id = id;
        scoreRec.m_score = score;
        scoreRec.m_bestScore = bestScore;
    }

    if (!ok)
    {
        if (pMsg) *pMsg = "Parse error";
        return false;
    }

    if (pScores) *pScores = std::move(scores);
    return true;
}

bool parseScores(vector<ScoreRec>* pScores,
                 string* pMsg)
{
    string scoresData;
    if (!readScores(&scoresData))
    {
        if (pMsg) *pMsg = "Error reading scores";
        return false;
    }

    return parseScores(scoresData, pScores, pMsg);
}
