#include "Common.hpp"
#include "Protocol.hpp"
#include "Problem.hpp"
#include "ParseUtils.hpp"
#include "StringUtils.hpp"
#include "FileUtils.hpp"

using std::string;

void usage()
{
    printf("Usage: getproblems <start> <end>\n");
}

int main(int argc, char *argv[])
{
    if (argc != 3)
    {
        usage();
        return 1;
    }

    uint64_t start = 0;
    uint64_t end = 0;
    if (!parseU64(argv[1], &start) ||
        !parseU64(argv[2], &end) ||
        start < 1 || end > 1000000 || start > end)
    {
        usage();
        return 1;
    }

    for (uint64_t id = start; id <= end; id++)
    {
        printf("%4" PRIu64 ": ", id);
        fflush(stdout);

        string existingProblemData;
        if (readProblem(id, &existingProblemData))
        {
            printf("exists\n");
            continue;
        }

        string msg;
        string problemData;
        if (!Protocol::getProblem(id, &problemData, &msg))
        {
            printf("%s\n", msg.c_str());
            continue;
        }

        if (!writeProblem(id, problemData))
        {
            printf("Error writing file\n");
            continue;
        }

        printf("ok\n");
    }

    return 0;
}
