#include "Common.hpp"
#include "Problem.hpp"
#include <cmath>
#include <limits>

using std::string;

int main(int argc, char *argv[])
{
    printf("        hole  edges  verts   epsilon  maxScore   width  height\n");
    printf("       -----  -----  -----  --------  --------  ------  ------\n");

    for (uint64_t id = 1; ; id++)
    {
        string problemData;
        if (!readProblem(id, &problemData))
        {
            break;
        }

        string msg;
        Problem problem;
        if (!parseProblem(problemData, &problem, &msg))
        {
            printf("Error parsing problem %" PRIu64 ": %s\n", id, msg.c_str());
            continue;
        }

        double maxScore = std::ceil(1000.0 * std::log2((double)(problem.m_vertices.size() + problem.m_edges.size() + problem.m_hole.size()) / 6.0));

        int64_t minX = std::numeric_limits<int64_t>::max();
        int64_t maxX = std::numeric_limits<int64_t>::min();
        int64_t minY = std::numeric_limits<int64_t>::max();
        int64_t maxY = std::numeric_limits<int64_t>::min();
        for (auto& corner : problem.m_hole)
        {
            minX = std::min(minX, corner.x);
            maxX = std::max(maxX, corner.x);
            minY = std::min(minY, corner.y);
            maxY = std::max(maxY, corner.y);
        }
        int64_t width = maxX - minX;
        int64_t height = maxY - minY;

        printf("%4" PRIu64 ":  %5" PRIuZ "  %5" PRIuZ "  %5" PRIuZ "  %8" PRIu64 "  %8.0f  %6" PRIi64 "  %6" PRIi64 "\n",
               id,
               problem.m_hole.size(),
               problem.m_edges.size(),
               problem.m_vertices.size(),
               problem.m_epsilon,
               maxScore,
               width,
               height);
    }

    return 0;
}
