#include "Common.hpp"
#include "Best.hpp"
#include "Scores.hpp"
#include <cmath>

using std::string;
using std::vector;

int main(int argc, char *argv[])
{
    string msg;
    vector<ScoreRec> scores;
    if (!parseScores(&scores, &msg))
    {
        printf("%s\n", msg.c_str());
        return 1;
    }

    double myTotal = 0.0;
    double maxTotal = 0.0;

    for (uint64_t id = 1; ; id++)
    {
        string problemData;
        if (!readProblem(id, &problemData))
        {
            break;
        }

        Problem problem;
        if (!parseProblem(problemData, &problem, &msg))
        {
            printf("%s\n", msg.c_str());
            return 1;
        }

        Best best;
        best.load(id);
        uint64_t score = best.getScore();

        printf("%4" PRIu64 "", id);
        if (score == (uint64_t)-1)
        {
            printf("         ");
        }
        else
        {
            printf("  %7" PRIu64 "", score);
        }

        if (id - 1 < scores.size())
        {
            auto& scoreRec = scores[id - 1];
            if (scoreRec.m_id != id)
            {
                printf("ID mismatch!");
            }
            else
            {
                if (scoreRec.m_score == (uint64_t)-1)
                {
                    printf("         ");
                }
                else
                {
                    printf("  %7" PRIu64 "", scoreRec.m_score);
                }

                uint64_t bestScore = scoreRec.m_bestScore;
                if (score != (uint64_t)-1)
                {
                    if (bestScore == (uint64_t)-1 || score < bestScore)
                    {
                        bestScore = score;
                    }
                }

                if (bestScore == (uint64_t)-1)
                {
                    printf("         ");
                }
                else
                {
                    printf("  %7" PRIu64 "", scoreRec.m_bestScore);
                }

                double myScore = 0.0;

                if (score == (uint64_t)-1)
                {
                    printf("         ");
                }
                else
                {
                    myScore = std::ceil(1000.0 * std::log2((double)(problem.m_vertices.size() * problem.m_edges.size() * problem.m_hole.size()) / 6.0) * std::sqrt((double)(bestScore + 1) / (double)(score + 1)));
                    myTotal += myScore;

                    printf("  %7.0f", myScore);
                }

                double maxScore = std::ceil(1000.0 * std::log2((double)(problem.m_vertices.size() * problem.m_edges.size() * problem.m_hole.size()) / 6.0));
                maxTotal += maxScore;

                printf("  %7.0f", maxScore);

                double percent = 0.0;
                if (maxScore != 0.0)
                {
                    percent = myScore / maxScore * 100.0;
                }
                printf("  %5.1f%%", percent);
            }
        }

        printf("\n");
    }

    double percentTotal = 0.0;
    if (maxTotal != 0.0)
    {
        percentTotal = myTotal / maxTotal * 100.0;
    }

    printf("                                 %7.0f  %7.0f  %5.1f%%\n",
           myTotal,
           maxTotal,
           percentTotal);

    return 0;
}
