#include "Common.hpp"
#include "Graphics.hpp"
#include "DrawProblem.hpp"
#include "Problem.hpp"
#include "Geometry.hpp"
#include "FastPoints.hpp"
#include "StringUtils.hpp"
#include <limits>

using std::string;
using std::vector;

int main(int argc, char *argv[])
{
    uint64_t id = 1;

    bool destroyRequested = false;
    bool destroyed = false;
    Graphics gr;

    auto onDestroy = [&]()
    {
        destroyed = true;
    };
    gr.setOnDestroy(onDestroy);

    auto onPaint = [&]()
    {
        gr.beginPaint(true);

        Problem problem;
        Solution solution;
        string text;

        string msg;
        if (!parseProblem(id, &problem, &msg))
        {
            text = strprintf("Problem %" PRIu64 ": %s", id, msg.c_str());
        }
        else
        {
            text = strprintf("Problem %" PRIu64 "", id);
            parseSolution(id, &solution);
        }

        vector<Point> points = getPointsInPolygon(problem.m_hole);

        FastPoints fastPoints;
        fastPoints.init(problem, points);

        drawProblem(gr, problem, fastPoints, points, solution, text);

        gr.endPaint();
    };
    gr.setOnPaint(onPaint);

    auto onKey = [&](Key key)
    {
        switch (key)
        {
        case Key::Esc:
        case Key::Q:
        case Key::ShiftQ:
        {
            if (!destroyRequested)
            {
                destroyRequested = true;
                gr.destroy();
            }
            break;
        }
        case Key::Right:
        {
            if (id <= 131)
            {
                id++;
                gr.requestPaint();
            }
            break;
        }
        case Key::Left:
        {
            if (id >= 2)
            {
                id--;
                gr.requestPaint();
            }
            break;
        }
        case Key::CtrlRight:
        {
            if (id <= 131)
            {
                if (id <= 122)
                {
                    id += 10;
                }
                else
                {
                    id = 132;
                }
                gr.requestPaint();
            }
            break;
        }
        case Key::CtrlLeft:
        {
            if (id >= 2)
            {
                if (id >= 11)
                {
                    id -= 10;
                }
                else
                {
                    id = 1;
                }
                gr.requestPaint();
            }
            break;
        }
        default: ;
        }
    };
    gr.setOnKey(onKey);

    string msg;
    if (!gr.create(1000, 1000, "Show Problem", &msg))
    {
        fprintf(stderr, "%s\n", msg.c_str());
        return 1;
    }

    while (!destroyed)
    {
        if (Graphics::update(true))
        {
            continue;
        }
    }

    return 0;
}
