#include "Common.hpp"
#include "Best.hpp"
#include "Scores.hpp"
#include "IsValid.hpp"
#include "CalcScore.hpp"
#include "Protocol.hpp"
#include "StringUtils.hpp"
#include "FileUtils.hpp"
#include <cmath>

using std::string;
using std::vector;

int main(int argc, char *argv[])
{
    string msg;

    printf("Updating scores...");
    fflush(stdout);

    string scoresData;
    if (!Protocol::getScores(&scoresData, &msg))
    {
        printf("\n");
        printf("%s\n", msg.c_str());
        return 1;
    }

    printf("\n");

    vector<ScoreRec> scores;
    if (!parseScores(scoresData, &scores, &msg))
    {
        printf("%s\n", msg.c_str());
        return 1;
    }

#if 1
    if (!writeScores(scoresData))
    {
        printf("Error writing scores\n");
        return 1;
    }
#endif

    for (uint64_t id = 1; ; id++)
    {
        string problemData;
        if (!readProblem(id, &problemData))
        {
            break;
        }

        printf("%4" PRIu64 ": ", id);
        fflush(stdout);

        Problem problem;
        if (!parseProblem(problemData, &problem, &msg))
        {
            printf("%s\n", msg.c_str());
            continue;
        }

        string solutionData;
        if (!readSolution(id, &solutionData))
        {
            printf("No solution\n");
            continue;
        }

        if (solutionData.empty())
        {
            printf("Empty solution\n");
            continue;
        }

        Solution solution;
        if (!parseSolution(solutionData, &solution, &msg))
        {
            printf("%s\n", msg.c_str());
            continue;
        }

        if (!isValid(problem, solution))
        {
            printf("Invalid solution\n");
            continue;
        }

        uint64_t score = calcScore(problem, solution);

        if (id - 1 >= scores.size())
        {
            printf("Best submitted score not available\n");
            continue;
        }

        auto& scoreRec = scores[id - 1];
        if (scoreRec.m_id != id)
        {
            printf("ID mismatch!");
            continue;
        }

        if (scoreRec.m_score != (uint64_t)-1)
        {
            if (score > scoreRec.m_score)
            {
                printf("Score is worse!\n");
                continue;
            }

            if (score == scoreRec.m_score)
            {
                printf("Score unchanged\n");
                continue;
            }
        }

        printf("Submitting... ");
        fflush(stdout);

#if 1
        string responseData;
        if (!Protocol::submitSolution(id, solutionData, &responseData, &msg))
        {
            printf("%s\n", msg.c_str());
            continue;
        }

        string responseFileName = strprintf("../poseids/%" PRIu64 "", id);
        if (!writeFile(responseFileName, responseData))
        {
            printf("Error writing response data\n");
            printf("%s\n", responseData.c_str());
            continue;
        }
#endif

        printf("ok\n");
    }

    return 0;
}
